;
        TITLE   FLEX_DRIVE
;
; FLEXDRV.SYS
; Flexible But Not Floppy
; Conventional Memory RAM Drive
; Version 3.00
;
; Copyright (C) 1989-1994,
; David Ryan Sheldon,
; All Rights Reserved
;
; Documentation appears at end of source
;
;-----------------------------------------------
;
        PAGE    198,81
;
BPT     EQU     BYTE PTR
WPT     EQU     WORD PTR
SHO     EQU     SHORT
OFS     EQU     OFFSET
;
;-----------------------------------------------
;
_TEXT   SEGMENT PARA PUBLIC 'CODE'
        ASSUME  CS:_TEXT
;
BSELBL  LABEL   BYTE
;
;-----------------------------------------------
;
; device header
;
        DD      -1            ;driver link
        DW      0             ;device attribute
        DW      STRAT         ;strategy routine
INTRPT  DW      INITD         ;interrupt rtne
        DB      1,7 DUP(0)    ;device name
;
;-----------------------------------------------
;
STRAT   PROC    FAR
;
; strategy routine
;
        MOV     CS:RHDOFS,BX  ;reg hdr ofs
        MOV     CS:RHDSEG,ES  ;req hdr seg
        RET
;
STRAT   ENDP
;
;-----------------------------------------------
;
INTRP   PROC    FAR
;
; interrupt routine dispatcher
;
        PUSH    SI
;
        DB      0BEh          ;MOV SI,iw
RHDSEG  DW      ?             ;req hdr segment
;
        PUSH    DS
        MOV     DS,SI         ;req hdr seg in DS
;
        PUSHF
;
        DB      0BEh          ;MOV SI,iw
RHDOFS  DW      ?             ;req hdr offset
;
        PUSH    AX
        CLD
        PUSH    BX
        INC     SI
        LODSW                 ;command + unit
        OR      AL,AL         ;unit 0 ?
        PUSH    BP
        XCHG    AL,AH         ;command in AL/AX
        PUSH    ES
        JNZ     ERROR0        ;no - unknown unit
;
        DEC     AX            ;media check ?
        JZ      MEDIA0
;
        DEC     AX            ;build BPB ?
        JZ      BUILD0
;
        SUB     AL,2          ;read ?
        JZ      RDWR00
;
        SHR     AL,1          ;ignore verify
        CMP     AL,2          ;write ?
        JZ      RDWR01
;
        CMP     AL,4          ;done / error ?
        JB      SDONE0
;
;- - - - - - - - - - - - - - - - - - - - - - - -
;
; errors
;
        MOV     AL,3          ;unknown command
        JMP SHO ERROR2
;
ERROR0: MOV     AL,1          ;unknown unit
        JMP SHO ERROR2
;
ERROR1: XOR     AX,AX
        MOV     [SI+0Fh],AX   ;adjusted count
        MOV     AL,8          ;sector not found
;
ERROR2: MOV     AH,81h        ;error + done
        JMP SHO SDONE1
;
;- - - - - - - - - - - - - - - - - - - - - - - -
;
; build BPB
;
BUILD0: MOV     [SI+11h],CS   ;local segment
;
        DB      0C7h,44h,0Fh  ;MOV WPT [SI+0Fh],
BPBARY  DW      BPBLCK        ;BPB pointer array
;
        JMP SHO SDONE0
;
;- - - - - - - - - - - - - - - - - - - - - - - -
;
; media check
;
MEDIA0: MOV BPT [SI+0Bh],1    ;media not changed
;
;- - - - - - - - - - - - - - - - - - - - - - - -
;
; set done
;
SDONE0: INC     AH            ;set done bit
;
SDONE1: MOV     [SI],AX       ;status word
        JMP SHO RDWR08
;
;- - - - - - - - - - - - - - - - - - - - - - - -
;
; read/write initialization
;
RDWR00: MOV     AX,0DDF6h     ;set read code
        JMP SHO RDWR02
;
RDWR01: MOV     AX,0C5FFh     ;set write code
;
RDWR02: MOV     CS:RWCB00,AL  ;set read/write
        MOV     CS:RWCB01,AH
;
        DB      0B8h          ;MOV AX,iw
CNFW00  DW      ?             ;total sectors
;
        LES     BX,[SI+0Fh]   ;ES=1st sec,BX=cnt
        MOV     BP,ES         ;BP = first sector
        SUB     AX,BP         ;first sec valid ?
        JBE     ERROR1        ;no - sec not fnd
;
        PUSH    CX
        MOV     CX,210h       ;2 X status word
        CMP     AX,BX         ;last sec valid ?
        JNC     RDWR03        ;yes
;
        MOV     [SI+0Fh],AX   ;adjusted count
        XCHG    AX,BX         ;set last valid
;
RDWR03: RCR     CX,1          ;set error bit
        MOV     [SI],CX       ;set status word
        SHR     CL,1          ;CL = 4
        LDS     SI,[SI+0Bh]   ;DS:AX = xfer ptr
        MOV     AX,SI         ;AX = xfer ofs
        SHR     AX,CL         ;remove low nybble
;
        DB      0B1h          ;MOV CL,ib
CNFB00  DB      ?             ;paras/sec expon
;
        SHL     BP,CL         ;cnvt sec to para
;
        DB      81h,0C5h      ;ADD BP,iw
CNFW01  DW      ?             ;drive base seg
;
        SHL     BX,CL         ;cnvt cnt to paras
        PUSH    DX
        MOV     DX,DS         ;xfer seg in DX
        ADD     DX,AX         ;adjust for ofs
        PUSH    DI
        MOV     DI,SI         ;DI = SI
;
        DB      31h           ;XOR gr,gr
RWCB00  DB      0F6h          ;gr = SI/DI
;
;- - - - - - - - - - - - - - - - - - - - - - - -
;
; main read/write loop
;
RDWR04: MOV     AX,0FFFh      ;max paras/pass
        SUB     BX,AX         ;have that many ?
        JA      RDWR05        ;yes - do them
;
        ADD     AX,BX         ;no - done ?
        JZ      RDWR07        ;yes - exit
;
        XOR     BX,BX         ;no - last pass
;
RDWR05: MOV     DS,DX         ;transfer pointer
        AND     SI,0Fh        ; in both
        MOV     ES,DX         ; DS:SI and
        AND     DI,0Fh        ; ES:DI
;
        DB      8Eh           ;MOV sr,BP
RWCB01  DB      0DDh          ;sr = DS/ES
;
        ADD     DX,AX         ;adj transfer seg
        ADD     BP,AX         ;adj drive seg
        SHL     AX,1          ;convert
        SHL     AX,1          ; paragraph count
        SHL     AX,1          ; to word count
        XCHG    AX,CX         ;in CX for REP
;
RDWR06: REP     MOVSW         ;transfer data
        JMP     RDWR04        ; and loop
;
;- - - - - - - - - - - - - - - - - - - - - - - -
;
; restore registers and exit
;
RDWR07: POP     DI
        POP     DX
        POP     CX
;
RDWR08: POP     ES
        POP     BP
        POP     BX
        POP     AX
        POPF
        POP     DS
        POP     SI
        RET
;
INTRP   ENDP
;
;-----------------------------------------------
;
; resident driver bios parameter block
;
BPBLCK  LABEL   BYTE
CNFW02  DW      20h           ;bytes/sector
CNFB01  DB      0             ;sectors/cluster
        DW      1             ;reserved sectors
CNFB02  DB      1             ;number of FAT's
CNFW03  DW      ?             ;directory entries
CNFW04  DW      ?             ;total sectors
        DB      0F8h          ;media descriptor
CNFW05  LABEL   WORD          ;sectors/FAT
        DB      ?             ; low byte
CNFB03  DB      ?             ; high byte
        DW      1             ;sectors/track
        DW      1             ;number of heads
        DW      0             ;hidden sectors
;
BPBLEN  EQU     $-OFS BPBLCK
;
; end of resident portion
;
;-----------------------------------------------
;
; beginning of boot sector
;
        ORG     ($+0Fh-BSELBL) AND 0FFF0h
;
RESEND  LABEL   BYTE          ;end resident code
;
        DB      0EBh,0FEh,90h ;DOS branch
        DB      'DRS  3.3'    ;OEM name
;
BPBDRV  LABEL   BYTE
;
;-----------------------------------------------
;
; drive formatter
;
FMREL   PROC    FAR
;
; formatter relocator routine
; (also moves volume label to root directory)
;
; IF = CLR !
; AX = 0
; BX = 1st FAT word (0FFF8h)
; DX = 2nd FAT word (0FFh or 0FFFFh)
; BP = clear dest offset for read/write loop
;
; DF = SET ! (down)
; CX = pre label move clearing word count
; DS = device driver segment
; SI = end of unexpanded volume label
; ES = relocation segment
; DI = pre label move clearing dest offset
;
; ON STACK ! (listed in POP order)
;
; for relocator:
; LABEL NAME + FORMT RELOCATION WORD COUNT
; FAR RETURN OFFSET FOR FMREL
; FAR RETURN SEGMENT FOR FMREL
;
; for relocated formatter:
; CLEAR SOURCE OFFSET FOR READ/WRITE LOOP
; POST BPB MOVE CLEARING WORD COUNT
; FAT SEGMENT
; FAT SEGMENT
; FAR RETURN OFFSET FOR FORMT
; FAR RETURN SEGMENT FOR FORMT
;
; clear out last three words of volume label
;  and, possibly, pre-relocation formatter
;
FMREL0: REP     STOSW
;
; expand rest of volume label
;  and move relocatable formatter
;
        MOVSW
        MOVSW
        MOV     CL,5
        REP     STOSW
        POP     CX            ;CX = 14h
        REP     MOVSW
;
; set up registers so the relocatable formatter
;  can copy the resident BPB to the boot sector
;
        CLD
        MOV     SI,OFS BPBLCK
        PUSH    CS
        MOV     CL,BPBLEN
        POP     ES
        MOV     DI,OFS BPBDRV
        RET
;
; IF = CLR !
; AX = 0
; BX = 1st FAT word (0FFF8h)
; DX = 2nd FAT word (0FFh or 0FFFFh)
; BP = clear dest offset for read/write loop
;
; DF = CLR ! (up)
; CX = BPB length (13h)
; DS = ES = driver load segment
; SI = resident BPB offset
; DI = drive BPB offset
;
; ON STACK ! (listed in POP order)
;
; CLEAR SOURCE OFFSET FOR READ/WRITE LOOP
; POST BPB MOVE CLEARING WORD COUNT
; FAT SEGMENT
; FAT SEGMENT
; FAR RETURN OFFSET FOR FORMT
; FAR RETURN SEGMENT FOR FORMT
;
FMREL   ENDP
;
;- - - - - - - - - - - - - - - - - - - - - - - -
;
; Notes:
;     The recocatABLE formatter is the following
; routine, at its original load address. The
; relocatED formatter is the same code, after it
; has been moved to where it will be executed.
; It is always executed from the very end of the
; FAT, just below the volume label (or first
; root directory entry).
;     Its length must not exceed 28 bytes, if 32
; byte sectors are to be supported, because 3 of
; the 4 bytes below it may become DOS signature
; FAT entries, and the read/write routine will
; clear out only even-numbered byte counts.
;     In the event that it is shorter than 32
; bytes, it is still kept as close to the root
; directory as possible. If the relocated
; segment were written in source, it would look
; something like this:
;
;       ORG     0
;0000:
;       DB      n DUP(?)      ;reserved, n=4 min
;XXXX:
;       FORMT   PROC          ;relocated
;       FORMT   ENDP          ; formatter
;0020:
;       DB      32 DUP(x)     ;volume label
;0040:
;       DB      32 DUP(0)     ;area after label
;0060:
;
;     The base offset of the FORMT routine is
; variable, 0004h minimum.
;     The base offset of the volume label is
; constant, 0020h.
;     Following the volume label is an area that
; is filled with at least 32 0's, whether it is
; the second root directory entry or the first
; data sector. This will serve as a source of
; 0's for the resident read/write loop to clear
; out the relocated formatter.
;     The only real trick here, is how the
; pre-relocation formatter gets cleared out. If
; the drive is small, its remnants may be above
; the label, in an uncleared area (it can't
; clear itself). If the drive is large, the
; remnants may be below the relocated formatter.
; And, of course, there are cases where it is
; partially overwritten by the label and/or
; relocated formatter. It is really simpler than
; it sounds. It is cleared either when the
; volume label is expanded, by the relocator, or
; when the relocator/relocatable formatter/label
; remnants are cleared, by the relocated
; formatter. This is done by comparing addresses
; and setting the counts for the two procedures,
; accordingly.
;     For the relocator, addresses for the end
; of the volume label and the end of the pre-
; relocation formatter remnants are calculated,
; and the higher of the two is used as the
; starting point (remember, a reverse move).
;     For the relocated formatter, addresses for
; the end of the pre-relocation formatter and
; itself are calculated, and the lesser of the
; two is used as the ending point (forward).
;
;- - - - - - - - - - - - - - - - - - - - - - - -
;
FORMT   PROC    FAR
;
; relocatable formatter routine
;
; IF = CLR !
; AX = 0
; BX = 1st FAT word (0FFF8h)
; DX = 2nd FAT word (0FFh or 0FFFFh)
; BP = clear dest offset for read/write loop
;
; DF = CLR ! (up)
; CX = BPB length (13h)
; DS = ES = driver load segment
; SI = resident BPB offset
; DI = drive BPB offset
;
; ON STACK ! (listed in POP order)
;
; CLEAR SOURCE OFFSET FOR READ/WRITE LOOP
; POST BPB MOVE CLEARING WORD COUNT
; FAT SEGMENT
; FAT SEGMENT
; FAR RETURN OFFSET FOR FORMT
; FAR RETURN SEGMENT FOR FORMT
;
; copy the resident BPB into the boot sector
;
        REP     MOVSB
        POP     SI            ;SI = 40h
;
; clear a predetermined numnber of words,
;  starting at the end of the BPB, continuing up
;  to an address calculated as the lesser of:
; 1) the end of the pre-relocation formatter
; 2) bottom of the relocated formatter
;
        POP     CX
        REP     STOSW
;
; two FAT's (if only one FAT, it's done twice)
;
        XCHG    AX,BX
        MOV     CL,2
;
FORMT0: POP     ES
        MOV     DI,BX
        STOSW
        XCHG    AX,DX
        STOSW
        XCHG    AX,DX
        LOOP    FORMT0
;
; set up registers so the device driver read/
;  write routine can clean out the formatter
;
        PUSH    CS
        POP     DS
        PUSH    CS
        POP     ES
        STI
        MOV     DI,BP         ;DI = 4
        MOV     CL,(OFS LABEL0+1-OFS FORMT)/2
        RET
;
; IF = SET !
; BX = 0 !
;
; DF = CLR !
; CX = clear word count for read/write loop
; SI = clear source offset for read/write loop
; DI = clear dest offset for read/write loop
; DS = ES = relocation segment
;
FORMT   ENDP
;
;- - - - - - - - - - - - - - - - - - - - - - - -
;
; volume label
;
LABEL0  DB      'RAM DRIVE '  ;first 10 bytes
LETTER  DB      20h,28h       ;letter, attribute
LADTTM  DW      1800h,1A61h   ;time, date
LABEND  LABEL   BYTE
;
;- - - - - - - - - - - - - - - - - - - - - - - -
;
FORPR   PROC    NEAR
;
; pre-relocation formatter
;
; Entry is via FORPR2
;
; DF = CLR !
; AX = DI = 0
; CL = 0
; CH = don't care
; BX = paragraphs to clear above FORPR (0 is OK)
; DX = first segment above FORPR
; DS = device driver segment
; ES = don't care
; BP = relocation segment
; SI = 2nd FAT segment
; FPRMB0 = 12/16 bit FAT (0/FF)
; FPRMB1 = pre label move clearing word count
; FPRMB3 = post BPB move clearing word count
; FPRMB2 = pre label move clearing dest offset
; FPRMW0 = 1st FAT segment
;
; zero out any or all of drive above this
;  routine (possibly large, multi-segment area),
;  starting with the first paragraph above the
;  end of the routine
;
        ASSUME  DS:_TEXT
;
FORPR0: MOV     BX,AX
;
FORPR1: MOV     ES,DX
        ADD     DX,CX
        SHL     CX,1
        SHL     CX,1
        SHL     CX,1
        REP     STOSW
;
FORPR2: MOV     CH,10h
        SUB     BX,CX
        JA      FORPR1
;
        ADD     CX,BX
        JNZ     FORPR0
;
; set up most of the registers for
;  relocator and relocated formatter
;
        STD
        MOV     ES,BP
    MOV BP,(OFS FORMT+21h-OFS LABEL0) AND 0FFFEh
        MOV     BX,0FFF8h     ;first FAT word
;
        DB      0BAh,0FFh     ;MOV DX,xxFFh
FPRMB0  DB      0             ;12/16 bit FAT
;
        DB      0B1h          ;MOV CL,ib
FPRMB1  DB      ?             ;pre label move clearing word count
;
        DB      0BFh          ;MOV DI,00xx
FPRMB2  DB      ?,0           ;pre label move clearing dest offset
;
; set up stack for relocator
;  and relocated formatter
;
; far return segment for FORMT
;
        PUSH    CS
;
; far return offset for FORMT
;
        MOV     AL,OFS RDWR06-OFS BSELBL
        PUSH    AX
;
; FAT segment
;
        CLI
        PUSH    SI
;
; FAT segment
;
        DB      0BEh          ;MOV SI,iw
FPRMW0  DW      ?             ;1st FAT segment
;
        PUSH    SI
        MOV     SI,OFS LABEND-2
;
; post BPB move clearing word count
;
        DB      0B0h          ;MOV AL,ib
FPRMB3  DB      ?             ;post BPB move clearing word count
;
        PUSH    AX
;
; source offset for read/write loop
;
        MOV     AL,40h
        PUSH    AX
;
; far return segment for FMREL
;
        PUSH    ES
;
; far return offset for FMREL
;
        MOV     AL,OFS FORMT+20h-OFS LABEL0
        PUSH    AX
;
; label name + FORMT relocation word count
;
        MOV     AL,(OFS LADTTM+1-OFS FORMT)/2
        PUSH    AX
        MOV     AL,AH
        JMP     FMREL0
;
        ASSUME  DS:NOTHING
;
; IF = CLR !
; AX = 0
; BX = 1st FAT word (0FFF8h)
; DX = 2nd FAT word (0FFh or 0FFFFh)
; BP = clear dest offset for read/write loop
;
; DF = SET ! (down)
; CX = pre label move clearing word count
; DS = device driver segment
; SI = end of unexpanded volume label
; ES = relocation segment
; DI = pre label move clearing dest offset
;
; ON STACK ! (listed in PUSH order)
;
; for relocated formatter:
; FAR RETURN SEGMENT FOR FORMT
; FAR RETURN OFFSET FOR FORMT
; FAT SEGMENT
; FAT SEGMENT
; POST BPB MOVE CLEARING WORD COUNT
; CLEAR SOURCE OFFSET FOR READ/WRITE LOOP
;
; for relocator:
; FAR RETURN SEGMENT FOR FMREL
; FAR RETURN OFFSET FOR FMREL
; LABEL NAME + FORMT RELOCATION WORD COUNT
;
FORPR   ENDP
;
FPREND  EQU     ($+0Fh-RESEND) AND 0FFF0h
FPRPAR  EQU     ($+0Fh-BSELBL)/10h
;
;-----------------------------------------------
;
; uninitialized variables
;
PRMW02  LABEL   WORD          ;break segment
PRMW03  EQU     PRMW02+2      ;bytes/FAT
;
; physical copyright notice
;
        DB      ' Copyright (C) 1989-1994, '
        DB      'David Ryan Sheldon, '
        DB      'All Rights Reserved '
;
;-----------------------------------------------
;
INITD   PROC    FAR
;
; main initialization routine
;
; save registers in same order as driver
;
        PUSH    SI
        PUSH    DS
        PUSHF
        PUSH    AX
        MOV     AX,3000h      ;get DOS ver
        PUSH    BX
        PUSH    BP
        PUSH    ES
        PUSH    CX
        PUSH    DX
        PUSH    DI
;
; get DOS Ver. to see if drive number is valid
;
        INT     21h
        MOV     SI,CS:RHDOFS
        MOV     DS,CS:RHDSEG
        CMP     AL,3
        JB      INIT00
;
; get drive number and set label letter
;
        MOV     AL,[SI+16h]
        ADD     AL,41h
        MOV     CS:LETTER,AL
;
; parse command line parameters
;
INIT00: PUSH    CS
        CLD
        POP     ES
        ADD     SI,4
        PUSH    DS
        PUSH    SI
        LDS     SI,[SI+0Eh]
;
INIT01: CALL    GETPD
        JZ      INIT04
;
        MOV     DI,OFS SCHARS
        MOV     CL,4
        REPNZ   SCASB
        JNZ     INIT01
;
INIT02: CALL    GETPD
        JB      INIT02
;
        JZ      INIT04
;
INIT03: CALL    CS:PDSPCH[DI]
        SAHF
        JA      INIT03
;
; 1) calculate/set default values as required
; 2) calculate/set configration variables
; 3) set numeric display values
;
        ASSUME  DS:_TEXT
;
INIT04: PUSH    CS
        POP     DS
;
; bytes per sector and
; paragraphs per sector exponent
;
        MOV     AX,CNFW02     ;bps
        MOV     DX,AX
        MOV     BX,-5
;
INIT05: INC     BX
        SHR     DX,1
        JNC     INIT05
;
        MOV     CNFB00,BL     ;p/s exponent
        MOV     DI,OFS DEND00
        CALL    NASCI
        XCHG    AX,CX         ;CX = bps
        MOV     BL,SPCADJ[BX-1] ;BX = spc adj
;
; sectors per cluster
;
        MOV     AL,CNFB01
        OR      AL,AL
        JNZ     INIT06
;
        MOV     AH,2
        DIV     CX
;
INIT06: XCHG    AX,BP         ;BP = spc
;
; data clusters
;
        MOV     AX,PRMW01
        OR      AX,AX
        JNZ     INIT07
;
        INC     DX
        DIV     CX
        DIV     BP
;
; adjust cluster size if required
;
INIT07: CMP     AX,BX
        JA      INIT08
;
        CMP     BP,1
        JZ      INIT08
;
        SHR     BP,1
        SHL     AX,1
        JMP     INIT07
;
; set data clusters and sectors
; per cluster display strings
;
INIT08: MOV     DI,OFS DEND03
        CALL    NASCI
        XCHG    AX,BP         ;BP = dcs
        MOV     CNFB01,AL     ;spc
        MOV     DI,OFS DEND01
        CALL    NASCI
;
; bytes per cluster
;
        MUL     CX
        MOV     DI,OFS DEND02
        CALL    NASCI
        SHR     DX,1
        RCR     AX,1
;
; data area size
;
        MUL     BP
        SHL     AX,1
        RCL     DX,1
        MOV     DI,OFS DEND12
        CALL    NASCI
;
; root directory sectors
; root directory entries
;
        XCHG    AX,BX
        MOV     AL,5
        XCHG    AX,CX
        SHR     AX,CL
        XCHG    AX,CX         ;CX = rde/sec
        MOV     AX,PRMW00
        OR      AX,AX
        JNZ     INIT14
;
        OR      DH,DH
        JZ      INIT09
;
        MOV     DX,0FFh
;
INIT09: MOV     AH,8
        XCHG    AX,BX
        DIV     BX
        OR      DX,DX
        JZ      INIT10
;
        INC     AX
;
INIT10: CMP     CL,1
        JA      INIT11
;
        DEC     AX
        JNZ     INIT12
;
        CMP     AL,LABEL0
        JZ      INIT12
;
        MOV     AL,2
        JMP SHO INIT16
;
INIT11: SHR     CX,1
        ADD     AX,CX
        SHL     CX,1
        NEG     CX
        AND     AX,CX
        NEG     CX
        OR      AX,AX
        JNZ     INIT13
;
INIT12: INC     AX
;
INIT13: MOV     DX,130h
        CMP     AX,DX
        JBE     INIT15
;
        XCHG    AX,DX
        JMP SHO INIT15
;
INIT14: MUL     CX
        OR      DX,DX
        JZ      INIT15
;
        MOV     AH,0FFh
;
INIT15: MOV     DX,BP         ;dcs
        ADD     DX,CX
        NEG     CX
        AND     DX,CX
        NEG     CX
        CMP     AX,DX
        JB      INIT16
;
        XCHG    AX,DX
;
INIT16: CWD
        MOV     CNFW03,AX     ;rde
        MOV     DI,OFS DEND08
        CALL    NASCI
        DIV     CX
        MOV     PRMW00,AX     ;rds
        MOV     DI,OFS DEND07
        CALL    NASCI
        MOV     BX,CX
        MOV     CL,5
        SHL     BX,CL         ;BX = bps
;
; root directory size
;
        MUL     BX
        MOV     DI,OFS DEND11
        CALL    NASCI
;
; bits per FAT entry
;
        MOV     AX,0Ch
        CMP     BP,0FF5h
        JB      INIT17
;
        MOV     AL,10h
        MOV BPT FPRMB0,0FFh   ;12/16 bit FAT
;
INIT17: CWD
        MOV     DI,OFS DEND04
        CALL    NASCI
;
; sectors per FAT
;
        SHR     AX,1
        SHR     AX,1
        INC     BP
        INC     BP
        MUL     BP
        DEC     BP
        DEC     BP
        SHR     DX,1
        RCR     AX,1
        ADC     AX,DX
        ADD     AX,BX
        DEC     AX
        NEG     BX
        AND     AX,BX
        NEG     BX
        DIV     BX
        MOV     CNFW05,AX
        MOV     DI,OFS DEND05
        CALL    NASCI
        XCHG    AX,CX         ;CX = spf
;
; number of FAT's
;
        MOV     AL,CNFB02
        CBW
        MOV     DI,OFS DEND06
        CALL    NASCI
        XCHG    AX,SI         ;SI = num FAT's
;
; total FAT size
;
        XCHG    AX,CX
        MUL     BX
        MOV     PRMW03,AX     ;save bytes/FAT
        MUL     SI
        MOV     DI,OFS DEND10
        CALL    NASCI
;
; total sectors
;
        DIV     BX            ;FAT(s)
        INC     AX            ;boot
        XCHG    AX,CX
        MOV     AL,CNFB01
        XOR     AH,AH
        MUL     BP
        ADD     AX,CX
        ADC     DX,0
        ADD     AX,PRMW00
        ADC     DX,0          ;root dir
        MOV     CNFW00,AX
        MOV     CNFW04,AX
        MOV     DI,OFS DEND09
        CALL    NASCI
;
; total drive size
;
        MOV     SI,DX
        MOV     CL,CNFB00
        ADD     CL,4
        XOR     CH,CH
;
INIT18: SHL     AX,1
        RCL     DX,1
        LOOP    INIT18
;
        MOV     DI,OFS DEND13
        CALL    NASCI
        MOV     BX,DX
        MOV     BP,AX         ;BX:BP = tds
;
; drive size test
;
        OR      SI,SI
        JNZ     INIT20
;
        MOV     CX,4
;
INIT19: SHR     DX,1
        RCR     AX,1
        LOOP    INIT19
;
        OR      DX,DX
        JNZ     INIT20
;
        MOV     DI,CS
        ADD     DI,(RESEND-BSELBL)/10h
        MOV     CNFW01,DI     ;drive base seg
        ADD     AX,DI
        JC      INIT20
;
        CMP     AH,96h
        JAE     INIT20
;
        MOV     PRMW02,AX
        ADD     AH,0Ah
        MOV     ES,AX
        MOV     AX,55AAh
        MOV     DI,0Eh
        CLI
        XCHG    AX,ES:[DI]    ;T in AX, 5A in X
        XCHG    AX,CX         ;T in CX, 0 in AX
        XCHG    AX,ES:[DI]    ;5A in AX, 0 in X
        XCHG    AX,CX         ;T in AX, 5A in CX
        XCHG    AX,ES:[DI]    ;T in X, 0 in AX
        STI
        PUSH    CS
        POP     ES
        XCHG    AX,CX         ;5A in AX, 0 in CX
        CMP     AX,55AAh
        JNZ     INIT20
;
        JCXZ    INIT23
;
INIT20: MOV     AX,OFS DSTR66
;
; display error message and abort exit
;
INIT21: MOV     SI,DMMODE
        MOV     DI,OFS IHERRR
        CMP     SI,OFS IHELPS
        JZ      INIT22
;
        MOV     SI,OFS IVERBO
        MOV     DI,OFS IERROR
;
INIT22: STOSW
        MOV     AX,OFS DSTR00
        STOSW
        MOV     AX,OFS DSTR68
        STOSW
        XOR     AX,AX
        STOSW
        CALL    IDISP
        POP     DI
        POP     ES
        MOV     AL,1
        STOSB
        ADD     DI,8
        XOR     AX,AX
        STOSB
        STOSW
        MOV     AX,CS
        STOSW
        JMP     RDWR07
;
; test the number of sectors per FAT
;
INIT23: MOV     AX,OFS DSTR67
        CMP     CL,CNFB03
        JNZ     INIT21
;
; drive sizes OK - continue with installation
;
        MOV WPT INTRPT,OFS INTRP
;
; total memory used
;
        XCHG    AX,BP
        MOV     DX,BX
        ADD     AX,OFS RESEND
        ADC     DX,CX
        ADD     AX,20h        ;+32 for DOS
        ADC     DX,CX         ; in display only
        MOV     DI,OFS DEND15
        CALL    NASCI
;
; total overhead
;
        MOV     AX,OFS RESEND+20h ;+32 for DOS
        CWD
        MOV     DI,OFS DEND14
        CALL    NASCI
;
; device driver address
;
        MOV     DX,CS
        MOV     DI,OFS DVAL16+1
        MOV     CX,4
;
INIT24: XOR     AX,AX
        PUSH    CX
        MOV     CL,4
;
INIT25: SHL     DX,1
        RCL     AX,1
        LOOP    INIT25
;
        POP     CX
        ADD     AL,90h
        DAA
        ADC     AL,40h
        DAA
        STOSB
        LOOP    INIT24
;
; display initialization screen
;
        MOV     SI,DMMODE
        CALL    IDISP
;
; set request header parameters
;
        POP     DI
        POP     ES
        MOV     AX,1
        STOSB
        ADD     DI,8
        STOSB
        DEC     AX
        STOSW
        MOV     AX,PRMW02
        STOSW
        XCHG    AX,BX
        MOV     AX,OFS BPBARY
        STOSW
        MOV     AX,CS
        STOSW
;
; AX = CS
; BX = break segment
;
; set format parameters and registers for FORPR
;
        MOV     DX,AX
        ADD     DX,FPRPAR     ;DX
        ADD     AX,(RESEND-BSELBL)/10h
        MOV     CL,4
        MOV     DI,CNFW02     ;bps
        SHR     DI,CL
        ADD     AX,DI
        MOV     FPRMW0,AX     ;1st FAT seg
        MOV     DI,PRMW03     ;bytes/FAT
        SHR     DI,CL
        CMP BPT CNFB02,1
        JZ      INIT26
;
        ADD     AX,DI
;
INIT26: MOV     SI,AX         ;SI
        ADD     AX,DI
        DEC     AX
        DEC     AX
        MOV     BP,AX         ;BP
        DEC     AX
        DEC     AX
        SUB     AX,CNFW01
        DEC     CX
        SHL     AX,CL
        ADD     AL,3
        MOV     DI,(FPREND-1Eh)/2
        CMP     AX,DI
        JB      INIT27
;
        XCHG    AX,DI
;
INIT27: MOV     FPRMB3,AL
        MOV     AX,DX
        SUB     AX,4
        SUB     AX,BP
        JNC     INIT28
;
        XOR     AX,AX
;
INIT28: SHL     AX,CL
        ADD     AL,3
        MOV     FPRMB1,AL
        SHL     AX,1
        ADD     AL,38h
        MOV     FPRMB2,AL
        XOR     AX,AX
        SUB     BX,DX
        JNC     INIT29
;
        MOV     BX,AX
;
INIT29: MOV     CL,AL
        MOV     DI,AX
        JMP     FORPR2
;
        ASSUME  DS:NOTHING
;
; FPRMB3 = post BPB move clearing word count
;
;xDF = CLR !
;xAX = DI = 0
;xCL = 0
;xCH = don't care
;xBX = paragraphs to clear above FORPR (0 is OK)
;xDX = first segment above FORPR
;xDS = device driver segment
;xES = don't care
;xBP = relocation segment
;xSI = 2nd FAT segment
;xFPRMB0 = 12/16 bit FAT (0/FF)
; FPRMB1 = pre label move clearing word count
; FPRMB2 = pre label move clearing dest offset
; FPRMB3 = post BPB move clearing word count
;xFPRMW0 = 1st FAT segment
;
INITD   ENDP
;
;-----------------------------------------------
;
PCLSZ   PROC    NEAR
;
; parse cluster size
;
        CALL    GETPN
        OR      DL,DL
        JZ      PCLSZ0
;
        MOV     BX,DX
        DEC     BX
        TEST    BX,DX
        JNZ     PCLSZ0
;
        MOV     CS:CNFB01,DL
        MOV     BX,OFS PDSPC1
;
PCLSZ0: MOV WPT CS:[BX],OFS PNEXT
;
PCLSZ1: RET
;
PCLSZ   ENDP
;
;-----------------------------------------------
;
PDATA   PROC    NEAR
;
; parse data clusters
;
        CALL    GETPN
        DEC     DX
        CMP     DX,1FDEh
        JAE     PCLSZ1
;
        INC     DX
        MOV     CS:PRMW01,DX
        MOV     BX,OFS PDSPC2
        JMP     PCLSZ0
;
PDATA   ENDP
;
;-----------------------------------------------
;
PFATS   PROC    NEAR
;
; parse number of FAT's
;
        CALL    GETPN
        DEC     DX
        CMP     DX,1
        JA      PCLSZ1
;
        INC     DX
        MOV     CS:CNFB02,DL
        MOV     BX,OFS PDSPC3
        JMP     PCLSZ0
;
PFATS   ENDP
;
;-----------------------------------------------
;
PROOT   PROC    NEAR
;
; parse root directory sectors
;
        CALL    GETPN
        DEC     DX
        CMP     DX,0FF5h
        JAE     PCLSZ1
;
        INC     DX
        MOV     CS:PRMW00,DX
        MOV     BX,OFS PDSPC5
        JMP     PCLSZ0
;
PROOT   ENDP
;
;-----------------------------------------------
;
PSESZ   PROC    NEAR
;
; parse sector size
;
        CALL    GETPN
        TEST    DX,3E0h
        JZ      PCLSZ1
;
        MOV     BX,DX
        DEC     BX
        TEST    BX,DX
        JNZ     PCLSZ1
;
        MOV     CS:CNFW02,DX
        MOV     BX,OFS PDSPC6
        JMP     PCLSZ0
;
PSESZ   ENDP
;
;-----------------------------------------------
;
PTERS   PROC    NEAR
;
; parse enable terse display mode
;
        MOV WPT CS:INORML,0
        JMP SHO PHELP1
;
PTERS   ENDP
;
;-----------------------------------------------
;
PVERB   PROC    NEAR
;
; parse enable verbose display mode
;
        MOV WPT CS:DMMODE,OFS IVERBO
        JMP SHO PHELP0
;
PVERB   ENDP
;
;-----------------------------------------------
;
PHELP   PROC    NEAR
;
; parse help display mode
;
        MOV WPT CS:DMMODE,OFS IHELPS
        MOV WPT CS:PDSPC0,OFS PNEXT
;
PHELP0: MOV WPT CS:PDSPC8,OFS PNEXT
;
PHELP1: MOV     BX,OFS PNEXT
        MOV     CS:PDSPC7,BX
        JMP     BX
;
PHELP   ENDP
;
;-----------------------------------------------
;
PPAUS   PROC    NEAR
;
        MOV BPT CS:CNFB04,0B4h ;MOV AH,i
        JMP SHO PNEXT0
;
PPAUS   ENDP
;
;-----------------------------------------------
;
PLABL   PROC    NEAR
;
; parse disable automatic volume labeling
;
        XOR     AX,AX
        MOV     DI,OFS LABEL0
        MOV     CL,8
        REP     STOSW
        MOV     CS:PDSPC4,OFS PNEXT
;
; subroutine "falls" through to PNEXT
;
PLABL   ENDP
;
;-----------------------------------------------
;
PNEXT   PROC    NEAR
;
; parse find next parameter
;
PNEXT0: CALL    GETPD
        JB      PNEXT0
;
PNEXT1: LAHF
        RET
;
PNEXT   ENDP
;
;-----------------------------------------------
;
GETPD   PROC    NEAR
;
; parse get parameter designator
;
;   Returns: If terminator character:
;              CF = 0, ZF = 1 (zero)
;
;            If designator character:
;              CF = 0, ZF = 0 (above)
;              DI = dispatch pointer
;
;            If not designator or terminator:
;              CF = 1, ZF = 0 (below)
;
        LODSB
        MOV     CX,15h
        MOV     DI,OFS DCHARS
        REPNZ   SCASB
        JNZ     GETPD2
;
        SUB     DI,OFS DCHARS+1
        JZ      GETPD1
;
        DEC     DI
        AND     DI,1Eh
;
GETPD1: OR      AL,AL
        RET
;
GETPD2: MOV     CL,5
        MOV     DI,OFS TCHARS
        REPNZ   SCASB
        JZ      GETPD3
;
        STC
;
GETPD3: RET
;
GETPD   ENDP
;
;-----------------------------------------------
;
GETPN   PROC    NEAR
;
; parse get numeric value
;
        XOR     DX,DX
;
GETPN0: CALL    GETPD
        JAE     PNEXT1
;
        CMP     AL,39h
        JA      GETPN0
;
        CMP     AL,30h
        JB      GETPN0
;
        CMP     DX,1998h
        JA      GETPN1
;
        AND     AX,0Fh
        SHL     DX,1
        ADD     AX,DX
        SHL     DX,1
        SHL     DX,1
        ADD     DX,AX
        JMP     GETPN0
;
GETPN1: MOV     DX,0FFFFh
        JMP     PNEXT0
;
GETPN   ENDP
;
;-----------------------------------------------
;
NASCI   PROC    NEAR
;
; binary numeric to ASCII string
;
; DX:AX = binary value (0 to FFFF:FFFF)
; ES:DI = pointer to last byte of dest string
;
        PUSHF
        STD
        OR      DX,DX
        CLI
        PUSH    AX
        PUSH    BX
        MOV     BX,0Ah
        JNZ     NASCI0
;
        CMP     AX,BX
        JB      NASCI9
;
NASCI0: PUSH    CX
        MOV     CX,0FFFFh
        PUSH    BP
        XOR     BP,BP
        PUSH    SI
        MOV     SI,CX
        PUSH    DX
;
NASCI1: CMP     BP,DX
        JB      NASCI2
;
        JA      NASCI3
;
        CMP     BX,AX
        JA      NASCI3
;
        JZ      NASCI4
;
NASCI2: SHL     BX,1
        RCL     BP,1
        JNC     NASCI1
;
NASCI3: RCR     BP,1
        RCR     BX,1
;
NASCI4: SUB     AX,BX
        SBB     DX,BP
        JNC     NASCI5
;
        ADD     AX,BX
        ADC     DX,BP
;
NASCI5: RCL     CX,1
        RCL     SI,1
        TEST    BL,3
        JNZ     NASCI6
;
        SHR     BP,1
        RCR     BX,1
        JMP     NASCI4
;
NASCI6: OR      AL,30h
        STOSB
        MOV     AX,0FFFFh
        MOV     BL,0Ah
        NOT     CX
        XOR     SI,AX
        JZ      NASCI8
;
NASCI7: MOV     DX,SI
        MOV     SI,AX
        XCHG    AX,CX
        JMP     NASCI1
;
NASCI8: CMP     CX,BX
        JAE     NASCI7
;
        POP     DX
        POP     SI
        POP     BP
        POP     AX
        XCHG    AX,CX
        OR      AL,AL
        JZ      NASCIA
;
NASCI9: OR      AL,30h
        STOSB
;
NASCIA: POP     BX
        POP     AX
        INC     DI
        POPF
        RET
;
NASCI   ENDP
;
;-----------------------------------------------
;
IDISN   PROC    NEAR
;
IDISN0: INT     29h
;
IDISN1: LODSB
        OR      AL,AL
        JNZ     IDISN0
;
        MOV     SI,BX
;
;- - - - - - - - - - - - - - - - - - - - - - - -
;
IDISP   PROC    NEAR
;
; display initialization screen
;
        LODSW
        OR      AX,AX
        XCHG    AX,SI
        XCHG    AX,BX
        JNZ     IDISN1
;
IDISP0:
;
CNFB04  DB      0C3h,1        ;RETN or MOV AH,i
;
        INT     16h
        JZ      IDISP0
;
        XOR     AH,AH
        INT     16h
        RET
;
IDISP   ENDP
;
;- - - - - - - - - - - - - - - - - - - - - - - -
;
IDISN   ENDP
;
;-----------------------------------------------
;
; parsing character search strings
;
DCHARS  DB  '?HhCcDdFfLlRrSsTtVvPp' ;designators
TCHARS  DB      0,0Dh,0Ah,3Bh,1Ah ;term chars
SCHARS  DB      20h,2Ch,2Dh,2Fh ;start chars
;
;-----------------------------------------------
;
; sectors per cluster adjust table
;
SPCADJ  DB      9,14h,29h,54h,0A9h
;
;-----------------------------------------------
;
; miscelleneous initialized variables
;
PRMW00  DW      0             ;root dir sectors
PRMW01  DW      0             ;data clusters
;
;-----------------------------------------------
;
; display value numeric strings
;
DVAL00  DB      2 DUP(20h)
DEND00  DB      20h,0
DVAL01  DB      2 DUP(20h)
DEND01  DB      20h,0Dh,0Ah,0
DVAL02  DB      4 DUP(20h)
DEND02  DB      20h,0Dh,0Ah,0
DVAL03  DB      3 DUP(20h)
DEND03  DB      20h,0Dh,0Ah,0
DVAL04  DB      20h
DEND04  DB      20h,0Dh,0Ah,0
DVAL05  DB      2 DUP(20h)
DEND05  DB      20h,0Dh,0Ah,0
DVAL06  LABEL   BYTE
DEND06  DB      20h,0Dh,0Ah,0
DVAL07  DB      3 DUP(20h)
DEND07  DB      20h,0Dh,0Ah,0
DVAL08  DB      5 DUP(20h)
DEND08  DB      20h,0Dh,0Ah,0
DVAL09  DB      6 DUP(20h)
DEND09  DB      20h,0Dh,0Ah,0
DVAL10  DB      4 DUP(20h)
DEND10  DB      20h,0
DVAL11  DB      6 DUP(20h)
DEND11  DB      20h,0
DVAL12  DB      9 DUP(20h)
DEND12  DB      20h,0
DVAL13  DB      9 DUP(20h)
DEND13  DB      20h,0
DVAL14  DB      2 DUP(20h)
DEND14  DB      20h,0
DVAL15  DB      5 DUP(20h)
DEND15  DB      20h,0
DVAL16  DB      '     :0000',0Dh,0Ah,0
;
;- - - - - - - - - - - - - - - - - - - - - - - -
;
; display text strings
;
DSTR00  DB      'FLEXDRV.SYS',0
DSTR01  DB      'Flexible But Not Floppy',0Dh
        DB      0Ah,'(C)opyright 1989-1994, '
        DB      'David R. Sheldon',0Dh,0Ah
        DB      'Ver 3.00',0
DSTR02  DB      '(Use "H" for Parameter List)'
DSTR03  DB      0Dh,0Ah,0
DSTR04  DB      '{Default Value or Method'
DSTR05  DB      '}',0Dh,0Ah,0
DSTR06  DB      '[Sx]'
DSTR07  DB      ' Bytes',0
DSTR08  DB      'Bits'
DSTR09  DB      ' per ',0
DSTR10  DB      20h
DSTR11  DB      'Cluster',0
DSTR12  DB      "'"
DSTR13  DB      's',0
DSTR14  DB      'Boot'
DSTR15  DB      20h
DSTR16  DB      'Sector',0
DSTR17  DB      '{64K'
DSTR18  DB      ' Data',0
DSTR19  DB      ': ',0
DSTR20  DB      '  x'
DSTR21  DB      ' = ',0
DSTR22  DB      '[Rx]'
DSTR23  DB      ' Root Directory',0
DSTR24  DB      ' Entries',0
DSTR25  DB      ' Area',0
DSTR26  DB      'Total',0
DSTR27  DB      ' Overhead',0
DSTR28  DB      ' Memory Used',0
DSTR29  DB      '[Fx]'
DSTR30  DB      ' Number of'
DSTR31  DB      20h
DSTR32  DB      'FAT',0
DSTR33  DB      ' Entry',0
DSTR34  DB      ' Drive'
DSTR35  DB      ' Size',0
DSTR36  DB      ' Address',0
DSTR37  DB      '[Cx]',0
DSTR38  DB      '[Dx]',0
DSTR39  DB      '[H|?]',0
DSTR40  DB      '1, 2, 4, 8, 16, '
DSTR41  DB      '32, 64, 128',0
DSTR42  DB      ', 256, '
DSTR43  DB      '512',0
DSTR44  DB      '32',0
DSTR45  DB      '{',0
DSTR46  DB      '1',0
DSTR47  DB      ' to ',0
DSTR48  DB      '8158',0
DSTR49  DB      '4085',0
DSTR50  DB      ' or 2',12 DUP(20h)
DSTR51  DB      0
DSTR52  DB      '[L] Dis',0
DSTR53  DB      '[V]'
DSTR54  DB      ' En'
DSTR55  DB      'ables ',0
DSTR56  DB      'Automatic Volume Labeling'
        DB      0Dh,0Ah,'[T]',0
DSTR57  DB      'Terse',0
DSTR58  DB      'Verbose',0
DSTR59  DB      'Help'
DSTR60  DB      ' Display Mode',0Dh,0Ah,0
DSTR61  DB      '[P] Pauses For Any Key',0
DSTR62  DB      ' Space',0
DSTR63  DB      '2K',0
DSTR64  DB      'Example:',0Dh,0Ah,'DEVICE=',0
DSTR65  DB      ' S32 C1 R2 D3 V',0Dh,0Ah,0
DSTR66  DB      'Drive Too Large or '
        DB      'Insufficient Memory.'
        DB      0Dh,0Ah,0
DSTR67  DB      'Sectors per FAT Exceeds 255.'
        DB      0Dh,0Ah,'Try 64 Byte Sectors or'
        DB      ' 4084 Clusters.',0Dh,0Ah,0
DSTR68  DB      ' Installation Aborted.'
        DB      0Dh,0Ah,0
;
;- - - - - - - - - - - - - - - - - - - - - - - -
;
        EVEN
;
; display mode pointer
;
DMMODE  DW      INORML        ;display mode
;
; normal mode display array
;
INORML  DW      DSTR03,DSTR00,DSTR51-5,DSTR01
        DW      DSTR51-3,DSTR02,DSTR03,DSTR51-6
        DW      DSTR07,DSTR09,DSTR16,DSTR19
        DW      DSTR51-4,DVAL00,DSTR03,DSTR51-5
        DW      DSTR07,DSTR09,DSTR11,DSTR19
        DW      DSTR51-2,DVAL02,DSTR23,DSTR24
        DW      DSTR19,DSTR51-1,DVAL08,DSTR03
        DW      DSTR51-9,DSTR18,DSTR25,DSTR21
        DW      DVAL12,DSTR07,DSTR03,DSTR51-2
        DW      DSTR26,DSTR28,DSTR21,DSTR51-4
        DW      DVAL15,DSTR07,DSTR03,0
;
; verbose/error mode display array
;
IVERBO  DW      DSTR03,DSTR00,DSTR51-5,DSTR01
        DW      DSTR51-3,DSTR02,DSTR03,DSTR51-6
        DW      DSTR07,DSTR09,DSTR16,DSTR19
        DW      DSTR51-4,DVAL00,DSTR03,DSTR51-3
        DW      DSTR15,DSTR13,DSTR09,DSTR11
        DW      DSTR19,DSTR51-4,DVAL01,DSTR51-5
        DW      DSTR07,DSTR09,DSTR11,DSTR19
        DW      DSTR51-2,DVAL02,DSTR51-9,DSTR18
        DW      DSTR10,DSTR13,DSTR19,DSTR51-3
        DW      DVAL03,DSTR51-5,DSTR08,DSTR32
        DW      DSTR33,DSTR19,DSTR51-5,DVAL04
        DW      DSTR51-7,DSTR15,DSTR13,DSTR09
        DW      DSTR32,DSTR19,DSTR51-4,DVAL05
        DW      DSTR51-7,DSTR30,DSTR12,DSTR19
        DW      DSTR51-6,DVAL06,DSTR23,DSTR15
        DW      DSTR13,DSTR19,DSTR51-3,DVAL07
        DW      DSTR23,DSTR24,DSTR19,DSTR51-1
        DW      DVAL08,DSTR51-10,DSTR26,DSTR15
        DW      DSTR13,DSTR19,DVAL09,DSTR03
        DW      DSTR51-8,DSTR14,DSTR21,DSTR51-7
        DW      DVAL00,DSTR07,DSTR03,DSTR51-5
        DW      DSTR26,DSTR31,DSTR35,DSTR21
        DW      DSTR51-5,DVAL10,DSTR07,DSTR03
        DW      DSTR51-4,DSTR23,DSTR21,DSTR51-3
        DW      DVAL11,DSTR07,DSTR03,DSTR51-9
        DW      DSTR18,DSTR25,DSTR21,DVAL12
        DW      DSTR07,DSTR03,DSTR51-3,DSTR26
        DW      DSTR34,DSTR21,DVAL13,DSTR07
        DW      DSTR03,DSTR03
IERROR  DW      DSTR51-5,DSTR26,DSTR27,DSTR21
        DW      DSTR51-7,DVAL14,DSTR07,DSTR03
        DW      DSTR51-2,DSTR26,DSTR28,DSTR21
        DW      DSTR51-4,DVAL15,DSTR07,DSTR03
        DW      DSTR00,DSTR36,DSTR21,DVAL16,0
;
; help mode display array
;
IHELPS  DW      DSTR03,DSTR00,DSTR51-5,DSTR01
        DW      DSTR51-6,DSTR04,DSTR03,DSTR06
        DW      DSTR09,DSTR16,DSTR03,DSTR20
        DW      DSTR41,DSTR42,DSTR51-8,DSTR45
        DW      DSTR44,DSTR05,DSTR37,DSTR15
        DW      DSTR13,DSTR09,DSTR11,DSTR03
        DW      DSTR20,DSTR40,DSTR03,DSTR51-8
        DW      DSTR51-8,DSTR45,DSTR43,DSTR07
        DW      DSTR09,DSTR11,DSTR05,DSTR38
        DW      DSTR18,DSTR10,DSTR13,DSTR03
        DW      DSTR20,DSTR46,DSTR47,DSTR48
        DW      DSTR51-8,DSTR17,DSTR62,DSTR05
        DW      DSTR22,DSTR15,DSTR13,DSTR03
        DW      DSTR20,DSTR46,DSTR47,DSTR49
        DW      DSTR51-3,DSTR45,DSTR46,DSTR33
        DW      DSTR09,DSTR63,DSTR18,DSTR05
        DW      DSTR29,DSTR12,DSTR03,DSTR20
        DW      DSTR46,DSTR50,DSTR51-12,DSTR45
        DW      DSTR46,DSTR05,DSTR52,DSTR55
        DW      DSTR56,DSTR54,DSTR57,DSTR60
        DW      DSTR53,DSTR58,DSTR60,DSTR39
        DW      DSTR54,DSTR59,DSTR61,DSTR03
        DW      DSTR03
IHERRR  DW      DSTR64,DSTR00,DSTR65,0
;
;-----------------------------------------------
;
; parse dispatch vector table
;
PDSPCH  LABEL   WORD
PDSPC0  DW      PHELP         ;help mode
PDSPC1  DW      PCLSZ         ;cluster size
PDSPC2  DW      PDATA         ;data clusters
PDSPC3  DW      PFATS         ;number of FAT's
PDSPC4  DW      PLABL         ;volume labeling
PDSPC5  DW      PROOT         ;root sectors
PDSPC6  DW      PSESZ         ;sector size
PDSPC7  DW      PTERS         ;terse mode
PDSPC8  DW      PVERB         ;verbose mode
PDSPC9  DW      PPAUS         ;pause for any key
;
;-----------------------------------------------
;
_TEXT   ENDS
;
        END
;
;-----------------------------------------------
;
;GENERAL DESCRIPTION
;
;     This virtual "disk" driver may be used to
; create practically any conceivable drive that
; will fit in conventional memory. The resident
; code size is only 256 bytes; I considered
; naming it SMALLDRV or TINYDRV, but it may be
; used to create some relatively large drives.
;     A bit technical to configure for the
; layman, FLEXDRV was written primarily for
; those who want to specify parameters such as
; sector size and sectors per cluster over the
; entire range allowed by DOS. The user can even
; choose to have two copies of the FAT, for
; those rare occasions where a floppy must be
; closely emulated.
;     Flexibility is the key here; If a user
; wants to conserve memory, sector sizes as
; small as 32 bytes may be selected, with some
; slight loss in speed due to DOS buffer
; inefficiency. Customization for large/small
; file sizes and/or numerous/sparce file
; quantities is possible with this driver.
;     An added advantage is speed; I went out of
; my way to make sure this driver was at least
; as fast as anyone elses. I'll bet it's faster
; than the one you're presently using, although
; it may require a special program to make any
; valid measurements. Simply put, writting very
; fast code is a habit with me. Memory accesses
; and branching are kept to a minimum throughout
; the read/write portion of the code. These are
; the most predominant code attributes that can
; eat up CPU throughput. In short - We be fast.
; Besides the data transfers themselves, I think
; the most sluggish part of the driver is saving
; and restoring the registers, a requirement
; that I haven't quite yet figured out how to
; avoid (a really neat trick).
;     Another thing I'd like to boast about is
; the fact that this driver is well-behaived.
; Any drive created is a valid DOS drive. If
; some requested sectors are invalid, the valid
; ones are still transfered - many drivers
; transfer none. Although rare or nonexistant,
; transfers of 64 Kbytes and larger are handled
; without segment wrapping and with a minimal
; number of passes through the loop. The driver
; checks the unit number - most drivers don't.
; All request commands are handled the way DOS
; expects them to be. The 7 bytes in the device
; header name, documented as reserved, are; Even
; Micrsoft drivers violate this rule. The only
; error not tested for is invalid request header
; structure length; I think some DOS versions
; are capable of creating errors here, so I
; intentionally ignore the specified request
; header length. These little subtleties are
; things I've learned about by writing numerous
; block and character device drivers for DOS.
; Because many are not documented anywhere,
; experience is the only available teacher.
;
;-----------------------------------------------
;
;USAGE
;
; DEVICE=p:\FLEXDRV.SYS ....
;
; p:\ represents the drive and path specifier to
; tell DOS where to find the file FLEXDRV.SYS
;
; Assuming FLEXDRV.SYS has been renamed FD.SYS,
;
; FD.SYS [Sx][Cx][Dx][Rx][Fx][L][T][V][H|?][P]
;
; Sx  Sector size (bytes per sector)
;
;     x = 32, 64, 128, 256, 512
;
;     default = 32
;
; Cx  Cluster size (sectors per cluster)
;
;     x = 1, 2, 4, 8, 16, 32, 64, 128
;
;     default = calculated so that each cluster
;     is 512 bytes, regardless of sector size
;
;     Note: Cluster size may be reduced if it
;     can be done without increasing FAT size.
;     If cluster size is adjusted, the number of
;     data clusters is also adjusted so that the
;     amount of data space remains unaltered.
;
; Dx  Data clusters
;
;     x = 1 to 8158, installation will be
;     aborted if drive size is too large
;
;     default = calculated so that there is
;     64 Kbytes of data space
;
; Rx  Root directory sectors
;
;     x = 1 to 4085
;     the root directory must have space for 1
;     to 4085 entries and each entry is 32 bytes
;
;     default = calculated so that there is 1
;     root directory entry for every 2 Kbytes of
;     data space, +/-1 root directory sector (if
;     automatic volume labeling is enabled, the
;     default may be adjusted so that there is
;     space for at least 1 file directory entry)
;
;     Note: Root directory sectors may be
;     reduced if there are more root directory
;     entries than there are data clusters.
;     (one additional entry is allowed for a
;     volume label)
;
; Fx  Number of FAT's
;
;     x = 1 or 2
;
;     default = 1
;
;     Note: This option is included only for
;     completeness and, normally, should not be
;     used
;
; L   Disables automatic volume labeling
;
;     default = automatic labeling enabled
;
; T   Enables the terse display mode; Nothing
;     is displayed unless the driver aborts.
;     This mode is nice once the drive
;     parameters have been adjusted to there
;     final values. If both terse and help
;     display modes are enabled, the help
;     message will be displayed. If both terse
;     and verbose display modes are enabled, the
;     verbose message will be displayed.
;
;     default = terse display mode disabled
;
; V   Enables the verbose display mode; Complete
;     drive statistics are displayed. This mode
;     is handy when making adjustments to drive
;     parameters. If both help and verbose
;     display modes are enabled, only the help
;     message will be displayed.
;
;     default = verbose display mode disabled
;
; H or ?   Enables the help display mode; A list
;     of the optional parameters is displayed.
;     If both help and verbose display modes are
;     enabled, only the help message will be
;     displayed.
;
;     default = help display mode disabled
;
; P   Causes a "Hit any key" pause before
;     returning execution to DOS. This allows
;     the user to read the drive statistics
;     before it is zipped of the screen by
;     installation of some other software.
;
; NOTES:
;
;  1) parameters may appear in any order
;  2) parameters are interpreted only the first
;   time they are encountered with a valid value
;   and repeated values for the same parameter
;   are ignored (S32 S64 yields 32 bytes per
;   sector)
;  3) invalid parameters (like A=1) are ignored
;  4) invalid parameter values (like S=511) are
;   ignored
;  5) upper and/or lower case letters may be
;   used  (s256 is the same as S256)
;  6) only letters, numbers, the question mark,
;   and the semicolon are interpreted; spaces,
;   commas, colons, equal signs, etc. are
;   optional between parameters, and are ignored
;  7) a semicolon terminates interpretation
;   altogether, and all after is ignored (this
;   allows remarks in the CONFIG.SYS file)
;  8) reserved sectors, sectors per track, and
;   number of heads are always set to 1, hidden
;   sectors is always set to 0, and the media
;   descriptor byte is always set to 0F8h
;  9) sectors per FAT and total sectors are
;   calculated automatically
;
; EXMPLES:
;
;FLEXDRV.SYS
; Creates the default drive and has:
;        32 bytes per sector
;        16 sectors per cluster
;       128 data clusters (64K data space)
;        32 root directory sectors (32 entries)
;         1 copy of the FAT
;           automatic volume labeling enabled
;
;FLEXDRV.SYS S512
; Creates a drive that has:
;       512 bytes per sector
;         1 sector per cluster
;       128 data clusters (64K data space)
;         2 root directory sectors (32 entries)
;         1 copy of the FAT
;           automatic volume labeling enabled
;
;FLEXDRV.SYS C 4
; Creates a drive that has:
;        32 bytes per sector
;         4 sectors per cluster
;       512 data clusters (64K data space)
;        32 root directory sectors (32 entries)
;         1 copy of the FAT
;           automatic volume labeling enabled
;
;FLEXDRV.SYS S=512, C=1, D=313, R=4, F=2, L
; Creates a drive that has:
;       512 bytes per sector
;         1 sector per cluster
;       313 data clusters (156.5K data space)
;         4 root directory sectors (64 entries)
;         2 copies of the FAT
;           automatic volume labeling disabled
; (This emulates a DOS 160K floppy)
;
;FLEXDRV.SYS S:512,C:1,D:351,R:4,F:2,L
; Creates a drive that has:
;       512 bytes per sector
;         1 sector per cluster
;       351 data clusters (175.5K data space)
;         4 root directory sectors (64 entries)
;         2 copies of the FAT
;           automatic volume labeling disabled
; (This emulates a DOS 180K floppy)
;
;FLEXDRV.SYS S512, C2, D315, R7, F2, L
; Creates a drive that has:
;       512 bytes per sector
;         2 sectors per cluster
;       315 data clusters (315K data space)
;         7 root directory sectors (112 entries)
;         2 copies of the FAT
;           automatic volume labeling disabled
; (This emulates a DOS 320K floppy)
;
;FLEXDRV.SYS/S512/C2/D354/R7/F2/L
; Creates a drive that has:
;       512 bytes per sector
;         2 sectors per cluster
;       354 data clusters (354K data space)
;         7 root directory sectors (112 entries)
;         2 copies of the FAT
;           automatic volume labeling disabled
; (This emulates a DOS 360K floppy)
;
;FLEXDRV.SYS D256 R7
; Creates a drive that has:
;        32 bytes per sector
;        16 sectors per cluster
;       256 data clusters (128K data space)
;         7 root directory sectors (112 entries)
;         1 copy of the FAT
;           automatic volume labeling enabled
;
;FLEXDRV.SYS S32C1D1R1L;smallest possible drive
; Creates a drive that has:
;        32 bytes per sector
;         1 sector per cluster
;         1 data cluster (32 bytes data space)
;         1 root directory sector (1 entry)
;         1 copy of the FAT
;           automatic volume labeling disabled
;           a remark demonstrated
;
;-----------------------------------------------
;
;DISPLAY SCREENS
;
;- - - - - - - - - - - - - - - - - - - -
;
; normal mode
;
;FLEXDRV.SYS     Flexible But Not Floppy
;(C)opyright 1989-1994, David R. Sheldon
;Ver 3.00   (Use "H" for parameter list)
;
;       Bytes per Sector:     512
;      Bytes per Cluster:     512
; Root Directory Entries:    4085
;
;          Data Area =     655360 Bytes
;  Total Memory Used =     655360 Bytes
;
;- - - - - - - - - - - - - - - - - - - -
;
; verbose mode
;
;FLEXDRV.SYS     Flexible But Not Floppy
;(C)opyright 1989-1994, David R. Sheldon
;Ver 3.00   (Use "H" for parameter list)
;
;       Bytes per Sector:     512
;    Sectors per Cluster:     128
;      Bytes per Cluster:   65536
;          Data Clusters:    8158
;     Bits per FAT Entry:      16
;        Sectors per FAT:     510
;        Number of FAT's:       2
; Root Directory Sectors:    4085
; Root Directory Entries:    6300
;          Total Sectors:   19135
;
;        Boot Sector =        512 Bytes
;     Total FAT Size =      32768 Bytes
;     Root Directory =     201600 Bytes
;          Data Area =     653312 Bytes
;   Total Drive Size =     653312 Bytes
;
;     Total Overhead =        288 Bytes
;  Total Memory Used =     653600 Bytes
;FLEXDRV.SYS Address =  0000:0000
;
;- - - - - - - - - - - - - - - - - - - -
;
; help mode
;
;FLEXDRV.SYS     Flexible But Not Floppy
;(C)opyright 1989-1994, David R. Sheldon
;Ver 3.00      {Default Value or Method}
;
;[Sx] Bytes per Sector
;  x = 32, 64, 128, 256, 512        {32}
;[Cx] Sectors per Cluster
;  x = 1, 2, 4, 8, 16, 32, 64, 128
;                {512 Bytes per Cluster}
;[Dx] Data Clusters
;  x = 1 to 8158        {64K Data Space}
;[Rx] Root Directory Sectors
;  x = 1 to 4085   {1 Entry per 2K Data}
;[Fx] Number of FAT's
;  x = 1 or 2                        {1}
;[L] Disables Automatic Volume Labeling
;[T] Enables Terse Display Mode
;[V] Enables Verbose Display Mode
;[H|?] Enables Help Display Mode
;[P] Pauses For Any Key
;
;Example:
;DEVICE=FLEXDRV.SYS S32 C1 R2 D3 V
;
;- - - - - - - - - - - - - - - - - - - -
;
; normal or verbose mode with an error
;
;FLEXDRV.SYS     Flexible But Not Floppy
;(C)opyright 1989-1994, David R. Sheldon
;Ver 3.00   (Use "H" for parameter list)
;
;       Bytes per Sector:     512
;    Sectors per Cluster:     128
;      Bytes per Cluster:   65536
;          Data Clusters:    8158
;     Bits per FAT Entry:      16
;        Sectors per FAT:     510
;        Number of FAT's:       2
; Root Directory Sectors:    4085
; Root Directory Entries:  153712
;          Total Sectors: 2471274
;
;        Boot Sector =        512 Bytes
;     Total FAT Size =      77824 Bytes
;     Root Directory =    4918784 Bytes
;          Data Area = 1259143168 Bytes
;   Total Drive Size = 1265292288 Bytes
;
;Sectors per FAT Exceeds 255.
;Try 64 Byte Sectors or 4084 Clusters.
;FLEXDRV.SYS Installation Aborted.
;
;- - - - - - - - - - - - - - - - - - - -
;
; help mode with an error
;
;FLEXDRV.SYS     Flexible But Not Floppy
;(C)opyright 1989-1994, David R. Sheldon
;Ver 3.00      {Default Value or Method}
;
;[Sx] Bytes per Sector
;  x = 32, 64, 128, 256, 512        {32}
;[Cx] Sectors per Cluster
;  x = 1, 2, 4, 8, 16, 32, 64, 128
;                {512 Bytes per Cluster}
;[Dx] Data Clusters
;  x = 1 to 8158        {64K Data Space}
;[Rx] Root Directory Sectors
;  x = 1 to 4085   {1 Entry per 2K Data}
;[Fx] Number of FAT's
;  x = 1 or 2                        {1}
;[L] Disables Automatic Volume Labeling
;[T] Enables Terse Display Mode
;[V] Enables Verbose Display Mode
;[H|?] Enables Help Display Mode
;[P] Pauses For Any Key
;
;Drive Too Large or Insufficient Memory.
;FLEXDRV.SYS Installation Aborted.
;
;-----------------------------------------------
;
;CLUSTERS TOO LARGE
; (adjusted)
;
;     If the number of data clusters is less
; than or equal to the appropriate word in the
; table, and the sectors per cluster is > 1, the
; sectors per cluster is divided by two and the
; number of clusters is multiplied by two. This
; test/adjust procedure is repeated until the
; test fails. Sector size remains unaltered.
; Notice that the procedure does not alter the
; amount of data space or the size of the drive
; in any way. The same amount of data space is
; simply available in smaller pieces, allowing
; for more efficient allocation of file space.
;     This avoids cluster sizes that are too
; large for the configuration. The advantage of
; using large clusters is the reduction in the
; size of the FAT. If decreasing the cluster
; size does not increase the FAT size, this
; advantage is nullified. Sector size is
; reguarded as sacred because the user may want
; to take advantege of small or large sector
; sizes. So, to reduce cluster size, the sectors
; per cluster value is adjusted. To maintain
; data area size, the number of data clusters is
; adjusted accordingly.
;     The procedure also avoids 64 Kbyte
; clusters because there is no configuration
; smaller than 640 Kbytes where they won't be
; adjusted out. Although the DOS kernel seems to
; accept 64 Kbyte clusters, CHKDSK chokes on it.
; I wouldn't be at all surprised to find other
; programs that have a hard time with them.
;     If the number of data clusters can be
; doubled without increasing the FAT size,
; sectors per cluster may be too large. If the
; entire FAT fills less than one half of a
; sector, this may be true. If the current FAT
; is larger than this, doubling its size will
; increase the number of sectors per FAT.
;     The number of FAT entries per sector for
; the first sector of the FAT is two less than
; the same value for any other sector, because
; DOS reserves the first two entries. If there
; are fewer than 4,085 data clusters, FAT
; entries are always 12 bits per entry, so the
; 16 bit case may be ignored here. For 32, 64,
; 128, 256, 512 byte sectors, the number of 12
; bit FAT entries per sector are 21, 42, 85,
; 170, 341 respectively. The table values are
; found by subtracting the two reserved by DOS
; and integer division by two of the result. If
; the number of data clusters (i.e. FAT entries)
; is less than or equal to this value, it may be
; doubled without increasing the number of
; sectors required to contain the FAT.
;     Simple, eh.
;
; This table exists in SPCADJ, above,
; and is only shown here as a remark.
;
;       DW      9             ;32 byte sector
;       DW      14h           ;64 byte sector
;       DW      29h           ;128 byte sector
;       DW      54h           ;256 byte sector
;       DW      0A9h          ;512 byte sector
;
;-----------------------------------------------
;
;TOO MANY ROOT DIRECTORY SECTORS
; (adjusted)
;
;     Oops! The user ACCIDENTLY requested more
; directory entries than data clusters. This may
; not be what really happened, but that's how I
; treat it. For every directory entry, except
; the volume label, at least one data cluster is
; consumed. That goes for subdirectories, too.
;     It would be a total waste of RAM to build
; a drive with more entries than the number of
; data clusters, plus one for a volume label. So
; I take that number, round it up to the nearest
; whole sector, and build the drive with that
; many root directory entries. The user will get
; "Insufficient Disk Space" before he/she gets
; "File Creation Error" (descriptive, and it
; sounds like a message from God), realize that
; the drive is too small for what they wanted to
; do, and never even care about how many root
; directory entries are available.
;     The created drive, somehow, consumes less
; memory than they expected, too! Cheap!
;
;-----------------------------------------------
;
;DRIVE TOO LARGE OR INSUFFICIENT MEMORY
; (aborted)
;
;     This one is easy. This device driver was
; designed for conventional memory. That means
; a 640 K limit, in case anyone has forgotten.
; There is no way to use all 640 K for a drive;
; The vector table, BIOS data area, and COMMAND.
; COM take up some space, not to mention the
; DOS kernel and I/O interface. Different
; versions of DOS consume different amounts of
; memory, but it is safe to assume that the
; driver will not be loaded below an address of
; 0070:0000. I have limited the upper address to
; 9600:0000, which leaves 40 K Bytes for DOS to
; place the buffers and COMMAND.COM, etc. The
; maximum data clusters and root directory
; entries values allowed are based on the 598.25
; K between these two addresses, after the 288
; bytes of overhead have been subtracted. By
; using large data clusters, drives much larger
; than 640 K can still be specified. I simply
; tell the user that his configuration is too
; large and abort the installation.
;     By the way, for systems with less than
; 640 K, I make sure there is memory 40 K above
; the drive by testing the last word of said
; memory.
;     Because I allow drive configurations to
; be created that are too large for DOS, there
; is still the possibility of DOS displaying
; the message "Configuration Too Large" and
; halting. I could reduce the memory limits and
; avoid this, but the limit changes with DOS
; versions, and I don't want to impose limits
; on users of the smaller versions.
;     If the drive is too large and there are
; too many sectors per FAT, only the drive too
; large error message is displayed, because it
; is easier for the user to understand.
;
;Drive Too Large or Insufficient Memory.
;FLEXDRV.SYS Installation Aborted.
;
;-----------------------------------------------
;
;TOO MANY SECTORS PER FAT (or TOO MANY CLUSTERS)
; (aborted)
;
;     I guess I'd have to call this one a DOS
; shortcoming. I hate to call it a bug, because
; DOS is what DOS is and it's the programmers'
; responsiblity to work with it, however it may
; be defined.
;     Even though the "sectors per FAT" entry in
; the boot sector BPB allows for a word value
; (i.e. 0-65535), DOS only examines the lower
; byte of this word. That means that a drive can
; have a maximum of only 255 sectors per FAT,
; regardless of what the other dimensional
; parameters may be.
;     Sectors per FAT is affected by two things;
; 1) Sector size, and 2) The number of data
; clusters. Small sectors means more sectors to
; handle the same number of FAT entries. More
; data clusters means more and/or larger FAT
; entries (if clusters > 4,084, FAT entries
; require 16 bits each, otherwise they require
; only 12 bits).
;     For drives smaller than 600 Kbytes, this
; is a rare occurance, but it is possible. The
; following table shows the bytes per sector and
; sectors per cluster combinations that are
; affected, and the maximum number of data
; clusters that can be accomodated with 255
; sector FAT's:
;
; Bytes/Sec      Secs/Clstr      Max Data Clstrs
;    32              1                4084
;    32              2                4084
;    32              4                4084
;    64              1                8158
;
;     A mathematically funny thing happens when
; you go to 4085 clusters because of the 16-bit
; entries; You need 256, 32 byte FAT sectors!
;     Because the data cluster limit for the 64
; bytes per sector case is the maximum possible
; in any configuration, I don't allow any value
; higher, and the problem never manifests.
;     All other combinations are limited as
; well, but the drive sizes at those limits are
; beyond the size of any drive that exists in
; 640 K conventional memory.
;     Notice that these would be inefficient
; configurations on any magnetic media, because
; the available media space would be eaten up by
; sector gaps. RAM drives don't need sector
; gaps, but DOS's block device interface was
; designed for magnetic media, and this is fair
; justification for the shortcoming. In any
; event, I suspect that this is the reason for
; most other drivers not supporting sector sizes
; smaller than 128 bytes.
;
; What to do ???
;
;     The "clusters too large" problem discussed
; above was easy; In fact, the adjustment
; actually helped the user by allowing for more
; efficient allocation of file space, and it
; didn't change the amount of available data
; space or the amount of memory consumed.
;     This problem is quite the opposite; No
; matter what steps are taken to decrease the
; number of sectors required by the FAT, the
; user ends up with less than he/she expected.
;
; 1) Larger sectors:  fewer clusters available
;                     and more memory used
; 2) Larger clusters: fewer clusters available
; 3) Limit clusters:  fewer clusters available
;                     and a smaller drive
; 4) Complain to MS:  they'll just make it worse
; 5) Abort install:   perhaps the best answer
;
;     Note: increasing either sector size or
; sectors per cluster may not cure the problem;
; an adjustment to both may be required
;     Note: reducing the number of clusters may
; affect the number of root directory entries
;
;     Aborting installation seems like the way
; to go. I'll even tell them what the problem is
; and give them a couple of options to resolve
; it. That way, if the user needs the data
; cluster quantity, it can be achieved by using
; 64 byte sectors, and if the small drive size
; is sought, I let them know what the magic
; number of clusters is.
;     If the drive is too large and there are
; too many sectors per FAT, only the drive too
; large error message is displayed, because it
; is easier for the user to understand.
;
;Sectors per FAT Exceeds 255.
;Try 64 Byte Sectors or 4084 Clusters.
;FLEXDRV.SYS Installation Aborted.
;
;-----------------------------------------------
;
;RELOCATABLE FORMATTER EVOLUTION
;
;     This part of the driver has always been an
; interesting programming challenge. Even though
; there is no requirement to clear the data area
; portion of the drive, I have always felt that
; it was sloppy to leave behind what looks like
; garbage. All of the RAM drives I've ever
; written leave the data area (as well as the
; unused portion of the FAT) neatly filled with
; 0's. The trick is to have a small chunk of
; code, where it won't overwrite itself, that
; formats the drive. When it's done, somebody
; has to clear him out, too. If at all possible,
; the code should not be moved above the break
; address, or at least, the end of the driver
; file itself, so that no limitation is placed
; on drive size. And we certainly wouldn't want
; to leave it resident, seeing as it would
; consume memory and never be executed again
; (some drivers do it that way!).
;     In my very first RAM drive, I placed three
; sections of code in what would later become
; the boot sector, FAT, and first data cluster.
; As each was executed, the prior one was over-
; written. A REP STOSW and a JMP instruction
; existed in the resident porton of code, that
; cleared the last executed portion. The JMP was
; the same one that lives at the beginning of
; the boot sector. I remember thinking I was
; pretty slick, but a little experience has
; proven me wrong.
;     In the next revision, I came up with one
; that is similar to the two that followed it. I
; now have one chunk of relocatable code, with
; the volume label and some parameters attached
; at the end. If the data area was at least
; twice the size of the formatter, it was moved
; to the base of the data area. If the data area
; was too small, I placed it above the break
; address. As it executed, the parameters were
; accessed with LODSW instructions, which told
; the formatter where to do what. The volume
; label was placed outside of the parameter
; stream, so the SI register had to be saved
; while it was moved to the root directory.
; Finally, if the formatter was in the data
; area, it was cleared by the read/write loop of
; the resident driver itself, by copying 0's
; from an already formatted area of the drive
; (cute, huh). If the formatter was above the
; drive, I only had to use the part of the
; resident code that restored the registers. I'm
; now a genius, right! Maybe not.
;     The next version was much the same, but
; the volume label was placed directly in the
; parameter stream, allowing it to be accessed
; without modification of the SI register. Also,
; the formatter was always placed at the base of
; the data area, and even if the drive was
; smaller than the formatter, I cleared at least
; its size above it, giving the resident read/
; write loop someplace to get the 0's. This made
; the code smaller because no decision had to be
; made about which technique would be used, and
; the formatter always used the read/write loop
; to clear some 0's. By this time, I felt I was
; a close personal programming friend of Jesus!
;     But, now, I talk to God about writting my
; software! The volume label was moved below the
; formatter. When the relocator code executes,
; it places the label right at its final resting
; place in the root directory, and the formatter
; doesn't have to mess with it at all. The
; formatter is now executed from what will be
; either the second, third, and forth root
; directory entries, or the base of the data
; area, or some combination thereof. Of course,
; I think this is about as cute as it gets, but
; then, that's what I thought about the first
; one. I wonder ... What's next? Maybe I should
; talk with Him more often, say ... daily. Maybe
; He'll give me some insight on how to write the
; formatter so small that it takes 0 bytes, and
; I can just leave it resident without consuming
; memory. Perhaps He'll just help me write the
; resident portion without saving the registers.
;     A couple of pitfalls accompany this code.
; First, the relocator code has to reside below
; the formatter itself, and must be smaller than
; a sector. When the formatter is relocated, it
; is always moved in the down direction. This
; means the direction flag is set. We would not
; want the relocator to overwrite the formatter
; before it's completely relocated, or itself
; before it's done executing. The other pitfall
; is, as previously mentioned, no matter how
; small the data area and root directory are, we
; must clear out enough 0's above the formatter
; for the read/write loop to clear out the
; formatter. In the previous version, it took 7
; paragraphs (112 bytes), but in the newer
; version, it only takes 5. If I could get the
; formatter down to 28 bytes, I'd place it in
; the last FAT sector, then this wouldn't be
; nessecary. Sounds like it's time for another
; chat with The Big Guru. Whatever the future
; holds, keep in mind that it has to work with
; any drive from the 4 sector, 32 bytes per
; sector model, all the way up to the big ones.
;     A 28 byte formatter may be possible, if I
; do a majority of the formatting from non-
; relocated code. It would only be responsible
; for clearing out where it was relocated from,
; writting the FAT signature entries, and
; setting up the registers for the read/write
; loop. I'm working on it...
;     ... And here it is. Those daily chats with
; T.B.G. paid off. After applying the tricks
; learned above (using the read/write routine to
; do the final clean-up and relocating the
; volume label directly to its home) and moving
; some things around, it really boiled down to
; shifting responsibilities among the different
; code portions until things fit. Another change
; was to place the formatter parameters on the
; stack, a thought I've had before, but I always
; worry about wearing out my welcome when it
; comes to stack space. One change that helped
; was to move the BPB placement code into the
; relocatable formatter, which had been
; performed late in the initialization code.
; Originally, I did it that way to remove the
; overhead from the relocator and formatter. As
; it turned out, it was more important to have
; the space available for the relocator. Once I
; made these adjustments, there was room for the
; relocator and formatter.
;     As I said, responsibilities were shifted
; among the different code portions. Also, I
; now use a pre-relocation formatter routine to
; clear out any and all of the drive above it.
; This made all of the remaining area that
; required clearing accessable from within a
; single segment, shrinking the code to do it.
; The following list shows the code portions in
; the order that they appear in the file. The
; numbers in parenthesis indicate execution
; order. Each is followed by a brief list of
; responsibilities.
;
; 1. (2) formatter relocator
;        a. clear out the pre-relocation
;           formatter
;        b. relocate the volume label and
;           relocatable formatter
;        c. set up registers for the relocated
;           formatter
;        d. branch to relocated formatter with a
;           RETF (far return)
; 2. (3) relocatable formatter
;        a. copy the resident BPB into the boot
;           sector (overwrites part/relocator)
;        b. clear out remnants of relocator and
;           original relocatable formatter or,
;           at least, up to just below the
;           relocated formatter, itself
;        c. write FAT signature entries
;        d. set up registers so read/write loop
;           can clear relocated formatter
;        e. branch to resident read/write loop
;           with another RETF
; 3. (n/a) volume label
; 4. (1) pre-relocation formatter
;        a. clear out drive above itself, if any
;           (possibly, a rather large area)
;        b. set up registers for the relocator
;        c. set up the stack with far return
;           addresses and format parameters
;        d. branch to relocator with a JMP
;
;     A few things to note. First, the relocated
; formatter is always at the very end of the
; FAT, just below the volume label. Next, the
; second root directory entry (or first data
; sector, if only one root directory entry) is
; used as the source for 0's when the resident
; read/write loop clears out the relocated
; formatter. Finally, interrupts are disabled
; just before all the stuff is pushed on the
; stack, and reenabled once most of it has been
; popped off. Disabling interrupts allows for
; 20 or so more bytes of stack space, because no
; interrupts can occur that might otherwise use
; it (a dirty little trick). Because the major
; portion of the drive has already been cleared
; out by the pre-relocation formatter, it
; doesn't take very long for the relocator and
; relocated formatter to execute, so interrupts
; aren't turned off for very long at all. This
; may not be neccessary, as the stack space is
; probably available anyways, but it provides a
; margin of safety that makes me comfortable.
;     T.B.G. says we're finally done improving
; this thing. I'm glad; I've been playing with
; it for years and it was starting to get old.
; The only way I could be any happier with it,
; is if I had done it this way in the first
; place. You see, I'm not really all that smart.
; I guess I'm one of those people who has to do
; it a few hundred times before I get it right.
; There is one thing in my defense; All of them
; worked. The tangible payoff is size. Each one
; was smaller than the previous. Even so, there
; was something else inside me that wouldn't let
; it rest until it was really slick, that served
; as the true inspiration, size be damned.
;
;-----------------------------------------------
;
;NOTES ON WHY TWO FAT's
;
;     On magnetic media, two copies of the FAT
; are maintained as a matter of reliability. I'm
; not really sure how valid that is, because DOS
; doesn't maintain two copies of the root (or
; any other) directory, and I'd hate to be the
; guy to try and recover data from a disk with
; a good FAT and no root directory. I suppose
; there is some justification, as it would be
; possible to at least put all the files back
; together. This isn't the first time the
; argument has been brought up; I've seen the
; same discussion in several computer magazines.
;     RAM drives are totally different. I can't
; overemphisize how reliable RAM is, in this
; respect. There is no need for two copies of
; the FAT! I suppose there is some rare instance
; where a person might want two. In actuality, I
; included this support, merely as a matter of
; completeness; If you want 2, you can have 2!
; But, perhaps the verbose display mode's Total
; FAT Size value will change your mind.
;
;-----------------------------------------------
;
;MORE NOTES ON CLUSTER QUANTITY AND FAT SIZE
;
;     Because the 32 bytes per sector cases,
; where clusters > 4084, cause an abort, there
; are only three remaining combinations where so
; many clusters may actually be created:
;
;    Bytes per Sector   Sectors per cluster
;           64                   1
;           64                   2
;          128                   1
;
;     An attempt with any other combination will
; cause the drive to be too large. These are the
; only cases where 16 bit FAT's will actually be
; created. 16 bit FAT's use more memory, and
; should be avoided!
;     As an example, let's consider a drive with
; 64 byte sectors, 1 sector per cluster, and
; 4084 data clusters. The FAT entries in this
; case are 12 bits each. DOS always adds two
; FAT entries itself.
;
; (4084 + 2) x 12 / 8 = 6129 bytes
; (8 = bits per byte)
;
;     The FAT always takes whole sectors, so we
; have to round up to the nearest whole sector.
; The entire FAT consumes 96 sectors, or 6144
; bytes.
;     To demonstrate the effect of 16 bit FAT
; entries, let's add 1 data cluster. That is
; the number where DOS switches over to 16 bit
; FAT entries, 4085.
;
; (4085 + 2) x 16 / 8 = 8174 bytes
;
;     Rounding up means the FAT consumes 128
; sectors, or 8192 bytes. To get one more 64
; byte cluster, we used 2048 bytes more FAT
; space. If you need more than 4084 clusters,
; fine. But if you can avoid it, do, and save
; your memory for something more worthwhile.
;     I seriously considered setting the maximum
; data clusters to 4084, but remember, I wanted
; the device driver to be flexible, so it is.
;
;-----------------------------------------------
;
;SECTOR SIZES NOT SUPPORTED
;
; 16 Bytes per Sector, and Smaller
;
;     Well, let's first discuss smaller. Sector
; sizes of 1, 2, 4, or 8 bytes could be created
; but, due to the organization of memory in
; Intel microprocessors, it would make the
; resident read/write code rather clumsy. Memory
; is handed out in 16 byte "chunks", called
; paragraphs. Also, the problems existing for 16
; byte sectors, about to be discussed, would
; cause many programs to go totally crazy.
;     16 bytes per sector would seem to be a
; very nice sector size, because sectors and
; paragraphs would be the same thing. If I wrote
; a driver specifically for 16 byte sectors, it
; would be simpler, smaller, and a little bit
; faster. To be completely honest, I've never
; tried it, to see how the DOS I/O interface
; would react. I haven't because I am certain
; that many programs would not run properly.
;     First, and most importantly, the boot
; sector contains a 19 byte area called a BPB,
; or BIOS Parameter Block. It starts at the 12th
; byte in the boot sector, so its end is the
; 30th byte. It contains information, from which
; a program can determine which sectors are used
; for what. Most programs that access the BPB,
; do so by reading sector 0 into a buffer, then
; reading information at the appropriate buffer
; offset. Even though the sector size is the
; very first value in the BPB, most (or all)
; programs weren't designed to test this value
; to know it needs to read an additional sector,
; so that it would have the complete BPB in the
; buffer. Now it accesses information in the
; buffer beyond the first sector (garbage), and
; gets confused about the drive parameters.
;     Similarly, each directory entry contains
; 32 bytes, and reading a single sector would
; get you only half of an entry. This problem
; may not manifest itself; If the program got
; past the boot sector problem, it would already
; know that each entry was 2 sectors long.
;     When considering these oddball dimensions,
; one must keep in mind that the whole ball of
; wax was designed for magnetic media. 16 bytes
; per sector would be a laughable format for any
; magnetic media, as most of the media would be
; consumed in sector gaps. Even though my
; background is in hardware, the electronic
; ciruitry that you see on a fixed or floppy
; drive is a subject where my knowledge is quite
; limited, and I don't know if the drive circuit
; could even work with such small sectors.
;
; 1024 Bytes per Sector, or Larger
;
;     Primarily, larger sectors are more
; efficient for larger drives, because the
; total number of sectors can be kept small, and
; a proportionatly larger amount of data is
; transfered in each sector (large drives are
; used for large amounts of information, if you
; have that much information, you probably want
; a sizable piece of it). Also, in the case of
; magnetic media, reduction of media space
; consumed in sector gaps would make more space
; available for data. Smaller sectors are more
; efficient for smaller drives, because the
; data space can be allocated in smaller pieces.
; Sector sizes of 1024 bytes or larger would not
; be very efficient for a relatively small drive
; like 600 K or smaller.
;     The BIOS ROM's in the original IBM PC do
; support 1 K sectors, a forward-looking feature
; that, to date, hasn't been taken advantage of.
; In current versions of DOS, however, 1 Kbyte
; sectors are not supported. I think it may be
; in the future, though, with hard drives and CD
; ROM's, etc. getting larger and larger.
;     As a matter of curiousity, I have tried
; this with a RAM drive, and DOS wasted no time
; telling me that "Sectors Too Large" wasn't
; going to work.
;     Since then, I did manage to get more data
; onto a 360 K floppy by using 1 K sectors and
; 5 sectors per track, reducing the number of
; sector gaps (400 K didn't seem unreasonable).
; Another way is to put 10 sectors on each track
; (reduced gap length). Hey, DOS went from 8
; sectors per track to 9 with version 2.00; Why
; can't I try 10? Both methods required a
; special device driver and format program to
; work with DOS, and the 1 K sector driver had
; to fool DOS by managing each 1 K floppy sector
; as two 512 byte DOS sectors. Puttering around
; like that is one way to learn what the books
; don't tell you. I never use the drivers, but
; my curiousity was satisfied and it was a good
; experience.
;
;-----------------------------------------------
;
;DISPLAYED TOTAL OVERHEAD CALCULATION
;
;     The resident portion of this device driver
; consumes 256 bytes. Whenever a block device is
; installed via CONFIG.SYS, DOS adds a 32 byte
; control block, immediately above the driver.
;     The "Total Overhead" value displayed in
; the verbose display mode screen, includes this
; 32 byte block, as it would not otherwise be
; consumed. This prevents discrepancies in the
; available memory reported by programs like
; CHKDSK or Norton's SYSINFO.
;     Depending on the LASTDRIVE value, and the
; unit number assigned to the drive, DOS may
; also use an additional 64 or 80 bytes to
; increment the maximum block device number, or
; drive letter. It does this because it has to
; have a place to record the current directory
; for each block device. Each actually uses 79
; bytes, but the conglomerate is always rounded
; up to the nearest whole paragraph. The default
; value for LASTDRIVE is E, meaning that space
; is already allocated for up to 5 drives.
;     As far as I know, there is no way to test
; to see if this adjustment happens, or what the
; current LASTDRIVE value is. That means that
; some memory actually consumed may not be
; reflected in the displayed total overhead
; value. If you come up short by 64 or 80 bytes,
; this is where it went. It can be avoided by
; setting LASTDRIVE in the CONFIG.SYS file to
; a value at least as high as the drive letter
; that will be assigned to this device. The
; memory is still used, but at least your
; numbers will jive.
;
;     Total Overhead =        288 Bytes
;
;-----------------------------------------------
;
; Dave Sheldon
; Mesa, Arizona
;
;-----------------------------------------------
;
